@@ -1,6 +1,7 @@
 #include "lexdcompiler.h"
 #include <unicode/unistr.h>
 #include <memory>
+#include <chrono>
 #include <lttoolbox/string_utils.h>
 
 using namespace icu;
@@ -93,13 +94,41 @@ UnicodeString LexdCompiler::printPattern(const pattern_element_t& pat)
   return ret;
 }
 
+UnicodeString LexdCompiler::printFilter(const tag_filter_t& filter)
+{
+  UnicodeString ret;
+  ret += '[';
+  for (auto& it : filter.pos()) {
+    if (ret.length() > 1) ret += ',';
+    ret += name(it);
+  }
+  for (auto& it : filter.neg()) {
+    if (ret.length() > 1) ret += ',';
+    ret += '-';
+    ret += name(it);
+  }
+  for (auto& op : filter.ops()) {
+    if (ret.length() > 1) ret += ',';
+    ret += op->sigil;
+    ret += '[';
+    int ln = ret.length();
+    for (auto& it : *op) {
+      if (ret.length() > ln) ret += ',';
+      ret += name(it);
+    }
+    ret += ']';
+  }
+  ret += ']';
+  return ret;
+}
+
 trans_sym_t LexdCompiler::alphabet_lookup(const UnicodeString &symbol)
 {
   if (!symbol.hasMoreChar32Than(0, symbol.length(), 1)) {
     return trans_sym_t((int)symbol.char32At(0));
   } else {
     UString temp;
-    temp.append(symbol.getBuffer(), symbol.length());
+    temp.append(symbol.getBuffer(), (unsigned int)symbol.length());
     alphabet.includeSymbol(temp);
     return trans_sym_t(alphabet(temp));
   }
@@ -168,7 +197,7 @@ LexdCompiler::finishLexicon()
       die("Lexicon '%S' is empty.", err(name(currentLexiconId)));
     }
     appendLexicon(currentLexiconId, currentLexicon);
-    
+
     currentLexicon.clear();
     currentLexicon_tags.clear();
   }
@@ -189,14 +218,14 @@ LexdCompiler::internName(const UnicodeString& name)
 string_ref
 LexdCompiler::checkName(UnicodeString& name)
 {
-  const static UnicodeString forbidden = " :?|()<>[]*+";
+  const static UString forbidden = u" :?|()<>[]*+";
   name.trim();
   int l = name.length();
   if(l == 0) die("Unnamed pattern or lexicon.");
 
-  for(const auto &c: char_iter(forbidden)) {
+  for(const auto &c: forbidden) {
     if(name.indexOf(c) != -1) {
-      die("Lexicon/pattern names cannot contain character '%C'", c[0]);
+      die("Lexicon/pattern names cannot contain character '%C'", c);
     }
   }
   return internName(name);
@@ -221,7 +250,7 @@ LexdCompiler::readTagFilter(char_iter& iter, UnicodeString& line)
   bool tag_nonempty = false;
   bool negative = false;
   vector<shared_ptr<op_tag_filter_t>> ops;
-  for(; iter != iter.end() && (*iter).length() > 0; ++iter)
+  for(; !iter.at_end() && (*iter).length() > 0; ++iter)
   {
     if(*iter == "]" || *iter == "," || *iter == " ")
     {
@@ -279,34 +308,41 @@ LexdCompiler::readTagFilter(char_iter& iter, UnicodeString& line)
   return tag_filter_t();
 }
 
+void
+LexdCompiler::appendSymbol(const UnicodeString& s, lex_token_t& tok)
+{
+  if (shouldCombine) {
+    tok.symbols.push_back(alphabet_lookup(s));
+  } else {
+    for (int c = 0; c < s.length(); c++) {
+      tok.symbols.push_back(alphabet_lookup(s[c]));
+    }
+  }
+}
+
 void
 LexdCompiler::readSymbol(char_iter& iter, UnicodeString& line, lex_token_t& tok)
 {
-  if (*iter == "\\") {
-    if (shouldCombine) {
-      tok.symbols.push_back(alphabet_lookup(*++iter));
+  if ((*iter).startsWith("\\")) {
+    if ((*iter).length() == 1) {
+      appendSymbol(*++iter, tok);
     } else {
-      ++iter;
-      for (int c = 0; c < (*iter).length(); c++) {
-        tok.symbols.push_back(alphabet_lookup((*iter)[c]));
-      }
+      appendSymbol((*iter).tempSubString(1), tok);
     }
+  } else if ((*iter).startsWith(":")) {
+    appendSymbol((*iter).tempSubString(1), tok);
   } else if (*iter == "{" || *iter == "<") {
     UChar end = (*iter == "{") ? '}' : '>';
     int i = iter.span().first;
-    for (; iter != iter.end() && *iter != end; ++iter) ;
+    for (; !iter.at_end() && *iter != end; ++iter) ;
 
     if (*iter == end) {
       tok.symbols.push_back(alphabet_lookup(line.tempSubStringBetween(i, iter.span().second)));
     } else {
       die("Multichar symbol didn't end; searching for %S", err(end));
     }
-  } else if (shouldCombine) {
-    tok.symbols.push_back(alphabet_lookup(*iter));
   } else {
-    for (int c = 0; c < (*iter).length(); c++) {
-      tok.symbols.push_back(alphabet_lookup((*iter)[c]));
-    }
+    appendSymbol(*iter, tok);
   }
 }
 
@@ -315,7 +351,7 @@ LexdCompiler::processRegexTokenSeq(char_iter& iter, UnicodeString& line, Transdu
 {
   bool inleft = true;
   vector<vector<lex_token_t>> left, right;
-  for (; iter != iter.end(); ++iter) {
+  for (; !iter.at_end(); ++iter) {
     if (*iter == "(" || *iter == ")" || *iter == "|" || *iter == "/") break;
     else if (*iter == "?" || *iter == "*" || *iter == "+")
       die("Quantifier %S may only be applied to parenthesized groups", err(*iter));
@@ -325,11 +361,11 @@ LexdCompiler::processRegexTokenSeq(char_iter& iter, UnicodeString& line, Transdu
     else if (*iter == "[") {
       ++iter;
       vector<lex_token_t> sym;
-      for (; iter != iter.end(); ++iter) {
+      for (; !iter.at_end(); ++iter) {
         if (*iter == "]") break;
         else if (*iter == "-" && !sym.empty()) {
           ++iter;
-          if (*iter == "]" || iter == iter.end()) {
+          if (*iter == "]" || iter.at_end()) {
             --iter;
             lex_token_t temp;
             readSymbol(iter, line, temp);
@@ -440,7 +476,7 @@ LexdCompiler::processRegexGroup(char_iter& iter, UnicodeString& line, Transducer
   ++iter; // initial slash or paren
   int state = start_state;
   vector<int> option_ends;
-  for (; iter != iter.end(); ++iter) {
+  for (; !iter.at_end(); ++iter) {
     if (*iter == "(") {
       state = trans->insertNewSingleTransduction(0, state);
       state = processRegexGroup(iter, line, trans, state, depth+1);
@@ -466,7 +502,7 @@ LexdCompiler::processRegexGroup(char_iter& iter, UnicodeString& line, Transducer
     trans->linkStates(it, state, 0);
   if ((depth > 0 && *iter == "/") || (depth == 0 && *iter == ")"))
     die("Mismatched parentheses in regex");
-  if (iter == iter.end())
+  if (iter.at_end())
     die("Unterminated regex");
   ++iter;
   if (depth > 0) {
@@ -510,9 +546,9 @@ LexdCompiler::processLexiconSegment(char_iter& iter, UnicodeString& line, unsign
     int state = processRegexGroup(iter, line, seg.regex, 0, 0);
     seg.regex->setFinal(state);
   }
-  if(iter == iter.end() && seg.regex == nullptr && seg.left.symbols.size() == 0)
+  if(iter.at_end() && seg.regex == nullptr && seg.left.symbols.size() == 0)
     die("Expected %d parts, found %d", currentLexiconPartCount, part_count);
-  for(; iter != iter.end(); ++iter)
+  for(; !iter.at_end(); ++iter)
   {
     if((*iter).startsWith(" ") || *iter == ']')
       break;
@@ -525,12 +561,13 @@ LexdCompiler::processLexiconSegment(char_iter& iter, UnicodeString& line, unsign
       --iter;
       tags_applied = true;
     }
-    else if(*iter == ":")
+    else if((*iter).startsWith(":"))
     {
       if(inleft)
         inleft = false;
       else
         die("Lexicon entry contains multiple colons");
+      if ((*iter).length() > 1) readSymbol(iter, line, seg.right);
     }
     else readSymbol(iter, line, (inleft ? seg.left : seg.right));
   }
@@ -563,9 +600,9 @@ LexdCompiler::readToken(char_iter& iter, UnicodeString& line)
 
   const UnicodeString boundary = " :()[]+*?|<>";
 
-  for(; iter != iter.end() && boundary.indexOf(*iter) == -1; ++iter);
+  for(; !iter.at_end() && boundary.indexOf(*iter) == -1; ++iter);
   UnicodeString name;
-  line.extract(begin_charspan.first, (iter == iter.end() ? line.length() : iter.span().first) - begin_charspan.first, name);
+  line.extract(begin_charspan.first, (iter.at_end() ? line.length() : iter.span().first) - begin_charspan.first, name);
 
   if(name.length() == 0)
     die("Symbol '%S' without lexicon name at u16 %d-%d", err(*iter), iter.span().first, iter.span().second-1);
@@ -585,7 +622,7 @@ LexdCompiler::readToken(char_iter& iter, UnicodeString& line)
   {
     iter++;
     begin_charspan = iter.span();
-    for(; iter != iter.end() && (*iter).length() > 0 && *iter != ")"; iter++)
+    for(; !iter.at_end() && (*iter).length() > 0 && *iter != ")"; iter++)
     {
       if((*iter).length() != 1 || !u_isdigit((*iter).charAt(0)))
         die("Syntax error - non-numeric index in parentheses: %S", err(*iter));
@@ -595,6 +632,7 @@ LexdCompiler::readToken(char_iter& iter, UnicodeString& line)
     if(iter.span().first == begin_charspan.first)
       die("Syntax error - missing index in parenthesis");
     part = (unsigned int)StringUtils::stoi(to_ustring(line.tempSubStringBetween(begin_charspan.first, iter.span().first)));
+    if (part == 0) die("Invalid column number (0)");
     ++iter;
   }
 
@@ -653,7 +691,7 @@ LexdCompiler::readPatternElement(char_iter& iter, UnicodeString& line)
     if(*iter == ":")
     {
       iter++;
-      if(iter != iter.end() && (*iter).length() > 0)
+      if(!iter.at_end() && (*iter).length() > 0)
       {
         if(boundary.indexOf(*iter) == -1)
         {
@@ -690,7 +728,7 @@ LexdCompiler::processPattern(char_iter& iter, UnicodeString& line)
   const UnicodeString modifier = "+*?";
   const UnicodeString decrement_after_token = token_boundary + "([]";
 
-  for(; iter != iter.end() && *iter != ')' && (*iter).length() > 0; ++iter)
+  for(; !iter.at_end() && *iter != ')' && (*iter).length() > 0; ++iter)
   {
     if(*iter == " ") ;
     else if(*iter == "|")
@@ -769,7 +807,7 @@ LexdCompiler::processPattern(char_iter& iter, UnicodeString& line)
       processPattern(iter, line);
       if(*iter == " ")
         *iter++;
-      if(iter == iter.end() || *iter != ")")
+      if(iter.at_end() || *iter != ")")
         die("Missing closing ) for anonymous pattern");
       ++iter;
       tag_filter_t filter;
@@ -879,6 +917,10 @@ LexdCompiler::processNextLine()
     UnicodeString name = line.tempSubString(8);
     finishLexicon();
     currentPatternId = checkName(name);
+    if (lexicons.find(currentPatternId) != lexicons.end()) {
+      die("The name '%S' cannot be used for both LEXICONs and PATTERNs.",
+          err(name));
+    }
     inPat = true;
   }
   else if(line.length() > 7 && line.startsWith("LEXICON "))
@@ -926,6 +968,10 @@ LexdCompiler::processNextLine()
         die("Multiple incompatible definitions for lexicon '%S'.", err(name));
       }
     }
+    if (patterns.find(currentLexiconId) != patterns.end()) {
+      die("The name '%S' cannot be used for both LEXICONs and PATTERNs.",
+          err(name));
+    }
     inLex = true;
     inPat = false;
   }
@@ -948,7 +994,7 @@ LexdCompiler::processNextLine()
   {
     char_iter iter = char_iter(line);
     processPattern(iter, line);
-    if(iter != iter.end() && (*iter).length() > 0)
+    if(!iter.at_end() && (*iter).length() > 0)
       die("Unexpected %S", err(*iter));
   }
   else if(inLex)
@@ -958,9 +1004,10 @@ LexdCompiler::processNextLine()
     for(unsigned int i = 0; i < currentLexiconPartCount; i++)
     {
       entry.push_back(processLexiconSegment(iter, line, i));
+      if (*iter == "]") die("Unexpected closing bracket.");
     }
     if(*iter == ' ') ++iter;
-    if(iter != iter.end())
+    if(!iter.at_end())
       die("Lexicon entry has '%S' (found at u16 %d), more than %d components", err(*iter), iter.span().first, currentLexiconPartCount);
     currentLexicon.push_back(entry);
   }
@@ -1149,6 +1196,7 @@ LexdCompiler::buildPattern(const pattern_element_t &tok)
   if(patternTransducers.find(tok) == patternTransducers.end())
   {
     if (verbose) cerr << "Compiling " << to_ustring(printPattern(tok)) << endl;
+    auto start_time = chrono::steady_clock::now();
     Transducer* t = new Transducer();
     patternTransducers[tok] = NULL;
     map<string_ref, unsigned int> tempMatch;
@@ -1158,13 +1206,27 @@ LexdCompiler::buildPattern(const pattern_element_t &tok)
       for(unsigned int i = 0; i < pat_untagged.second.size(); i++)
       {
         auto pat = pat_untagged;
-        for(auto &pair: pat.second)
-        {
-          if(!pair.tag_filter.combine(tok.tag_filter.neg()))
-            die("Incompatible tag filters.");
+        bool taggable = true;
+        for (unsigned int j = 0; j < pat.second.size(); j++) {
+          auto& pair = pat.second[j];
+          if(!pair.tag_filter.combine(tok.tag_filter.neg())) {
+            taggable = false;
+            if (verbose) {
+              cerr << "Warning: The tags of " << to_ustring(printPattern(tok));
+              cerr << " conflict with " << to_ustring(printPattern(pat_untagged.second[j]));
+              cerr << " on line " << pat.first << "." << endl;
+            }
+          }
         }
-        if(!pat.second[i].tag_filter.combine(tok.tag_filter.pos()))
-          die("Incompatible tag filters.");
+        if(!pat.second[i].tag_filter.combine(tok.tag_filter.pos())) {
+          taggable = false;
+          if (verbose) {
+            cerr << "Warning: The tags of " << to_ustring(printPattern(tok));
+            cerr << " conflict with " << to_ustring(printPattern(pat_untagged.second[i]));
+            cerr << " on line " << pat.first << "." << endl;
+          }
+        }
+        if (!taggable) continue;
 
         matchedParts.clear();
         lineNumber = pat.first;
@@ -1184,8 +1246,12 @@ LexdCompiler::buildPattern(const pattern_element_t &tok)
       cerr << " is empty." << endl;
     }
     patternTransducers[tok] = t;
-    if (verbose)
-      cerr << "Done with " << to_ustring(printPattern(tok)) << endl;
+    if (verbose) {
+      auto end_time = chrono::steady_clock::now();
+      chrono::duration<double> diff = end_time - start_time;
+      cerr << "Done compiling " << to_ustring(printPattern(tok));
+      cerr << " in " << diff.count() << " seconds." << endl;
+    }
   }
   else if(patternTransducers[tok] == NULL)
   {
@@ -1249,6 +1315,7 @@ LexdCompiler::buildPatternWithFlags(const pattern_element_t &tok, int pattern_st
   if(patternTransducers.find(tok) == patternTransducers.end())
   {
     if (verbose) cerr << "Compiling " << to_ustring(printPattern(tok)) << endl;
+    auto start_time = chrono::steady_clock::now();
     Transducer* trans = (shouldHypermin ? hyperminTrans : new Transducer());
     patternTransducers[tok] = NULL;
     unsigned int transition_index = 0;
@@ -1298,13 +1365,20 @@ LexdCompiler::buildPatternWithFlags(const pattern_element_t &tok, int pattern_st
           }
           else
           {
-            if(i == idx)
-            {
-              if(!cur.tag_filter.combine(tok.tag_filter.pos()))
-                die("Incompatible tag filters.");
+            if (i == idx && !cur.tag_filter.combine(tok.tag_filter.pos())) {
+              if (verbose) {
+                cerr << "Warning: The tags of " << to_ustring(printPattern(tok));
+                cerr << " conflict with " << to_ustring(printPattern(pat.second[i]));
+                cerr << " on line " << pat.first << "." << endl;
+              }
+            }
+            if (!cur.tag_filter.combine(tok.tag_filter.neg())) {
+              if (verbose) {
+                cerr << "Warning: The tags of " << to_ustring(printPattern(tok));
+                cerr << " conflict with " << to_ustring(printPattern(pat.second[i]));
+                cerr << " on line " << pat.first << "." << endl;
+              }
             }
-            if(!cur.tag_filter.combine(tok.tag_filter.neg()))
-              die("Incompatible tag filters.");
           }
 
           Transducer* t;
@@ -1463,7 +1537,12 @@ LexdCompiler::buildPatternWithFlags(const pattern_element_t &tok, int pattern_st
         cerr << "FIXME" << endl;
       }
     }
-    if (verbose) cerr << "Done with " << to_ustring(printPattern(tok)) << endl;
+    if (verbose) {
+      auto end_time = chrono::steady_clock::now();
+      chrono::duration<double> diff = end_time - start_time;
+      cerr << "Done compiling " << to_ustring(printPattern(tok));
+      cerr << " in " << diff.count() << " seconds." << endl;
+    }
     patternTransducers[tok] = trans;
   }
   else if(patternTransducers[tok] == NULL)
